use std::env;
use std::fs::{read_to_string, File};
use std::io::prelude::*;
use std::path::Path;

use axum::{
    http::StatusCode,
    routing::{get, post},
    Json, Router,
};
use serde::{Deserialize, Serialize};
use spaceapi::Status;

mod status;

// TODO: Figure out where to ackshully put this.
const STATUS_FILE: &str = "status.json";

#[tokio::main]
async fn main() {
    if env::var("consumer_key").is_err() {
        panic!("env var consumer_key must be set");
    }
    if env::var("consumer_secret").is_err() {
        panic!("env var consumer_secret must be set");
    }

    let app = Router::new()
        .route("/status.json", get(root))
        .route("/api/update", post(the_doors));

    let listener = tokio::net::TcpListener::bind("0.0.0.0:3000").await.unwrap();

    println!("Start serving");
    axum::serve(listener, app).await.unwrap();
}

// Initialize empty status.
fn init_status() -> Status {
    let new_doors = status::TheDoors {
        aerie: false,
        cellar: false,
    };
    let (sensors, state) = status::update(new_doors);
    status::status(sensors, state)
}

// check given secret
// https://www.youtube.com/watch?v=aHKWVLH-ibY
fn auth(p: &TheDoors) -> bool {
    let consumer_secret = env::var("consumer_secret").unwrap();
    let consumer_key = env::var("consumer_key").unwrap();

    p.consumer_secret == consumer_secret && p.consumer_key == consumer_key
}

// Write status to file and return JSON string.
fn write_status(s: Status) -> String {
    let s = serde_json::to_string(&s).unwrap();
    let mut file = File::create(STATUS_FILE).unwrap();
    file.write_all(s.as_bytes()).unwrap();
    s
}

// Just output the current file. We assume it to be consistent.
// It may cease to or not yet exist. Then create an initial status and persist.
async fn root() -> String {
    if std::path::Path::new(STATUS_FILE).exists() {
        read_to_string(STATUS_FILE).unwrap_or(String::from("KAPOTT"))
    } else {
        write_status(init_status())
    }
}

// Input type for the API: Both fields are optional.
#[derive(Serialize, Deserialize)]
struct TheDoors {
    aerie: Option<bool>,
    cellar: Option<bool>,
    consumer_key: String,
    consumer_secret: String,
}

// The door can see through your soul.
// https://www.youtube.com/watch?v=bDQDp00oTP4
async fn the_doors(Json(payload): Json<TheDoors>) -> StatusCode {
    if !auth(&payload) {
        return StatusCode::FORBIDDEN;
    }

    let status: Status = if Path::new(STATUS_FILE).exists() {
        let contents = read_to_string(STATUS_FILE).expect("FCKAFD");
        serde_json::from_str(&contents).unwrap_or_else(|_| init_status())
    } else {
        init_status()
    };

    if let Some(sens) = status.sensors {
        // Get the current status as read from the file.
        let mut doors = sens.door_locked.into_iter();
        let aerie_current = doors
            .find(|d| d.metadata.location == "aerie")
            .unwrap_or(status::get_aerie(false));
        let cellar_current = doors
            .find(|d| d.metadata.location == "cellar")
            .unwrap_or(status::get_cellar(false));

        // New door status: fall back to respective current value if no update
        // is provided.
        let aerie = match payload.aerie {
            Some(s) => s,
            None => aerie_current.value,
        };
        let cellar = match payload.cellar {
            Some(s) => s,
            None => cellar_current.value,
        };

        let new_doors = status::TheDoors { aerie, cellar };
        let (sensors, state) = status::update(new_doors);

        let s = status::status(sensors, state);
        let s = serde_json::to_string(&s).unwrap();

        let mut file = File::create(STATUS_FILE).unwrap();
        file.write_all(s.as_bytes()).unwrap();
    }

    StatusCode::CREATED
}
